import numpy as np
cimport numpy as np

ctypedef double Real
from cython cimport boundscheck

cdef extern from "LineBP.h":
    void _LagrangeBP(int n, Real xi, Real *N, Real *dN) except +
    void _LagrangeGaussLobattoBP(int n, Real xi, Real *eps, Real *N, Real *dN) except +



@boundscheck(False)
def LagrangeBP_(int C, Real xi):

    cdef int n = C+2
    cdef np.ndarray[Real,ndim=1, mode='c']  N = np.zeros(n,dtype=np.float64)
    cdef np.ndarray[Real,ndim=1, mode='c'] dN = np.zeros(n,dtype=np.float64)

    _LagrangeBP(n, xi, &N[0], &dN[0])

    return N, dN


@boundscheck(False)
def LagrangeGaussLobattoBP_(int C, Real xi, np.ndarray[Real,ndim=1, mode='c'] eps):

    cdef int n = C+2
    cdef np.ndarray[Real,ndim=1, mode='c']  N  = np.zeros(n,dtype=np.float64)
    cdef np.ndarray[Real,ndim=1, mode='c'] dN  = np.zeros(n,dtype=np.float64)

    _LagrangeGaussLobattoBP(n, xi, &eps[0], &N[0], &dN[0])

    return N, dN


from __future__ import print_function
import gc, os, sys, inspect
from copy import deepcopy
from warnings import warn
from time import time
import numpy as np
import scipy as sp
from Florence.Utils import insensitive

from Florence.FiniteElements.Assembly import Assemble
from Florence.PostProcessing import *
from Florence.Solver import LinearSolver, FEMSolver

__all__ = ["TractionBasedStaggeredSolver"]

class TractionBasedStaggeredSolver(FEMSolver):
    """Staggered electromechanical solver based on
        traction/residual transmision
    """

    def __init__(self,**kwargs):
        super(TractionBasedStaggeredSolver, self).__init__(**kwargs)

        self.electric_dofs = None
        self.mechanical_dofs = None
        self.columns_in_mech = None
        self.columns_in_mech_idx = None
        self.columns_in_electric = None
        self.columns_in_electric_idx = None
        self.columns_out_mech = None
        self.columns_out_mech_idx = None
        self.columns_out_electric = None
        self.columns_out_electric_idx = None

    def GetBoundaryInfo(self,K,boundary_condition,formulation):

        all_dofs = np.arange(K.shape[0])
        self.electric_dofs = all_dofs[formulation.nvar-1::formulation.nvar]
        self.mechanical_dofs = np.array([],dtype=np.int64)
        self.mechanical_dofs = np.setdiff1d(all_dofs,self.electric_dofs)

        # GET BOUNDARY CONDITON FOR THE REDUCED MECHANICAL SYSTEM
        self.columns_in_mech = np.intersect1d(boundary_condition.columns_in,self.mechanical_dofs)
        self.columns_in_mech_idx = np.in1d(self.mechanical_dofs,boundary_condition.columns_in)

        # GET BOUNDARY CONDITON FOR THE REDUCED ELECTROSTATIC SYSTEM
        self.columns_in_electric = np.intersect1d(boundary_condition.columns_in,self.electric_dofs)
        self.columns_in_electric_idx = np.in1d(self.electric_dofs,boundary_condition.columns_in)


        # GET FREE MECHANICAL DOFs
        self.columns_out_mech = np.intersect1d(boundary_condition.columns_out,self.mechanical_dofs)
        self.columns_out_mech_idx = np.in1d(self.mechanical_dofs,boundary_condition.columns_out)

        # GET FREE ELECTROSTATIC DOFs
        self.columns_out_electric = np.intersect1d(boundary_condition.columns_out,self.electric_dofs)
        self.columns_out_electric_idx = np.in1d(self.electric_dofs,boundary_condition.columns_out)


        self.applied_dirichlet_mech = boundary_condition.applied_dirichlet[np.in1d(boundary_condition.columns_out,self.columns_out_mech)]
        self.applied_dirichlet_electric = boundary_condition.applied_dirichlet[np.in1d(boundary_condition.columns_out,self.columns_out_electric)]


        # MAPPED QUANTITIES
        # all_dofs = np.arange(0,K.shape[0])
        out_idx = np.in1d(all_dofs,boundary_condition.columns_out)
        idx_electric = all_dofs[formulation.nvar-1::formulation.nvar]
        idx_mech = np.setdiff1d(all_dofs,idx_electric)

        self.all_electric_dofs = np.arange(int(K.shape[0]/formulation.nvar))
        self.electric_out = self.all_electric_dofs[out_idx[idx_electric]]
        self.electric_in = np.setdiff1d(self.all_electric_dofs,self.electric_out)

        self.all_mech_dofs = np.arange(int(K.shape[0]/formulation.nvar*formulation.ndim))
        self.mech_out = self.all_mech_dofs[out_idx[idx_mech]]
        self.mech_in = np.setdiff1d(self.all_mech_dofs,self.mech_out)




    def Solve(self, formulation=None, mesh=None,
        material=None, boundary_condition=None,
        function_spaces=None, solver=None):
        """Main solution routine for FEMSolver """


        # CHECK DATA CONSISTENCY
        #---------------------------------------------------------------------------#
        function_spaces, solver = self.__checkdata__(material, boundary_condition, formulation, mesh, function_spaces, solver)
        #---------------------------------------------------------------------------#
        caller = inspect.getouterframes(inspect.currentframe(), 2)[1][3]
        if caller != "Solve":
            self.PrintPreAnalysisInfo(mesh, formulation)
        #---------------------------------------------------------------------------#


        # INITIATE DATA FOR NON-LINEAR ANALYSIS
        NodalForces, Residual = np.zeros((mesh.points.shape[0]*formulation.nvar,1),dtype=np.float64), \
            np.zeros((mesh.points.shape[0]*formulation.nvar,1),dtype=np.float64)
        # SET NON-LINEAR PARAMETERS
        self.NRConvergence = { 'Increment_'+str(Increment) : [] for Increment in range(self.number_of_load_increments) }

        # ALLOCATE FOR SOLUTION FIELDS
        TotalDisp = np.zeros((mesh.points.shape[0],formulation.nvar,self.number_of_load_increments),dtype=np.float64)

        # PRE-ASSEMBLY
        print('Assembling the system and acquiring neccessary information for the analysis...')
        tAssembly=time()

        # APPLY DIRICHELT BOUNDARY CONDITIONS AND GET DIRICHLET RELATED FORCES
        boundary_condition.GetDirichletBoundaryConditions(formulation, mesh, material, solver, self)

        # ALLOCATE FOR GEOMETRY - GetDirichletBoundaryConditions CHANGES THE MESH
        # SO EULERX SHOULD BE ALLOCATED AFTERWARDS
        Eulerx = np.copy(mesh.points)
        Eulerp = np.zeros((mesh.points.shape[0]))

        # FIND PURE NEUMANN (EXTERNAL) NODAL FORCE VECTOR
        NeumannForces = boundary_condition.ComputeNeumannForces(mesh, material, function_spaces)

        # ASSEMBLE STIFFNESS MATRIX AND TRACTION FORCES
        K, TractionForces = Assemble(self, function_spaces[0], formulation, mesh, material,
            Eulerx, Eulerp)[:2]

        print('Finished all pre-processing stage. Time elapsed was', time()-tAssembly, 'seconds')

        TotalDisp = self.StaggeredSolver(function_spaces, formulation, solver,
                K,NeumannForces,NodalForces,Residual,
                mesh,TotalDisp,Eulerx,Eulerp,material, boundary_condition)


        return self.__makeoutput__(mesh, TotalDisp, formulation, function_spaces, material)





    def StaggeredSolver(self, function_spaces, formulation, solver, K,
            NeumannForces,NodalForces,Residual,
            mesh,TotalDisp,Eulerx,Eulerp,material, boundary_condition):

        Tolerance = self.newton_raphson_tolerance
        LoadIncrement = self.number_of_load_increments
        LoadFactor = 1./LoadIncrement
        AppliedDirichletInc = np.zeros(boundary_condition.applied_dirichlet.shape[0],dtype=np.float64)

        # GET BOUNDARY CONDITIONS INFO FOR EACH INDIVIDUAL PROBLEM
        self.GetBoundaryInfo(K,boundary_condition,formulation)

        # SOLVE THE FIRST MECHANICAL PROBLEM
        nodal_forces_mech = np.zeros((self.mechanical_dofs.shape[0],1))
        residual_mech_neumann = -LoadFactor*NeumannForces[self.mechanical_dofs,:]
        residual_mech = residual_mech_neumann - self.ApplyDirichlet(K[self.mechanical_dofs,:][:,self.mechanical_dofs],
            nodal_forces_mech, self.mech_out, self.mech_in,self.applied_dirichlet_mech, LoadFactor=LoadFactor)

        dUm = self.SolveMechanics(mesh, formulation, solver, K, residual_mech, LoadFactor, initial_solution=True)

        # INITIATE TRANSMITTIVE FORCES
        self.force_up = np.zeros(formulation.ndim*mesh.points.shape[0])
        self.force_pu = np.zeros(mesh.points.shape[0])

        # INITIATE ELECTRIC NODAL FORCES AND RESIDUAL
        nodal_forces_electric = NodalForces[self.electric_dofs]

        for Increment in range(LoadIncrement):

            t_increment = time()

            # UPDATE FIXED DOFs FOR ELECTROSTATICS
            DeltaF_electric = LoadFactor*NeumannForces[self.electric_dofs]
            nodal_forces_electric += DeltaF_electric
            residual_electric = -self.ApplyDirichlet(K[self.electric_dofs,:][:,self.electric_dofs],
                nodal_forces_electric, self.electric_out, self.electric_in,self.applied_dirichlet_electric,LoadFactor)

            # COMPUTE FORCE TO BE TRANSMITTED TO ELECTROSTATIC
            K_pu = K[self.electric_dofs,:][:,self.mechanical_dofs]
            self.force_pu = K_pu.dot(dUm.flatten())

            # STORE PREVIOUS ELECTRIC POTENTIAL AT THIS INCREMENT
            Eulerp_n = np.copy(Eulerp)
            # APPLY INCREMENTAL POTENTIAL FOR FIXED POTENTIAL DoFs
            applied_dirichlet_electric_inc = LoadFactor*self.applied_dirichlet_electric
            Eulerp[self.electric_out] += applied_dirichlet_electric_inc

            # GET ONLY NORM OF FIXED DOFs (THAT IS WHERE RESIDUAL FORCES GET GENERATED)
            if Increment==0:
                self.NormForces = la.norm(residual_electric)
            # AVOID DIVISION BY ZERO
            if np.abs(self.NormForces) < 1e-14:
                self.NormForces = 1e-14

            Ke = deepcopy(K)

            residual_mech_from_elec = np.zeros((self.mechanical_dofs.shape[0],1))
            # DeltaF_mech = LoadFactor*NeumannForces[self.mechanical_dofs]
            # nodal_forces_mech += DeltaF_mech

            Iter = 0
            # ENTER NEWTON-RAPHSON FOR ELECTROSTATICS
            while np.abs(la.norm(residual_electric[self.electric_in])/self.NormForces) > Tolerance:

                # SOLVE ELECTROSTATIC PROBLEM ITERATIVELY
                dUe = self.SolveElectrostatics(Ke, residual_electric, formulation, solver, Iter)

                # UPDATE EULERIAN POTENTIAL - GET ITERATIVE ELECTRIC POTENTIAL
                Eulerp += dUe

                # RE-ASSEMBLE - COMPUTE INTERNAL TRACTION FORCES FOR ELECTROSTATICS
                Ke, TractionForces = Assemble(self, function_spaces[0], formulation, mesh, material,
                    Eulerx,Eulerp)[:2]

                # FIND THE ITERATIVE RESIDUAL
                residual_electric[self.electric_in] = TractionForces[self.columns_in_electric] - nodal_forces_electric[self.electric_in]

                # traction_forces_mech += TractionForces[self.mechanical_dofs]
                residual_mech_from_elec += TractionForces[self.mechanical_dofs] - nodal_forces_mech

                self.NRConvergence['Increment_'+str(Increment)] = np.append(self.NRConvergence['Increment_'+str(Increment)],\
                    np.abs(la.norm(residual_electric[self.electric_in])/self.NormForces))

                print('Iteration number', Iter, 'for load increment', Increment, 'with a residual of \t\t', \
                    np.abs(la.norm(residual_electric[self.electric_in])/self.NormForces))

                # UPDATE ITERATION NUMBER
                Iter +=1

                if Iter==self.maximum_iteration_for_newton_raphson or self.NRConvergence['Increment_'+str(Increment)][-1] > 500:
                    self.newton_raphson_failed_to_converge = True
                    break
                if np.isnan(np.abs(la.norm(residual_electric[self.electric_in])/self.NormForces)):
                    self.newton_raphson_failed_to_converge = True
                    break

            if self.newton_raphson_failed_to_converge:
                print("Solver blew up! Norm of incremental solution is too large")
                solver.condA = np.NAN
                Increment = Increment if Increment!=0 else 1
                TotalDisp = TotalDisp[:,:,:Increment]
                self.number_of_load_increments = Increment
                return TotalDisp


            # COMPUTE FORCE TO BE TRANSMITTED TO MECHANICS
            K_up = Ke[self.mechanical_dofs,:][:,self.electric_dofs]
            dUe = Eulerp - Eulerp_n
            self.force_up = K_up.dot(dUe)
            # self.force_up = K_up.dot(dUe[:,None])
            # print(self.force_up)

            # if Increment>0:
                # nodal_forces_mech = np.zeros_like(nodal_forces_mech)
                # residual_mech = residual_mech_neumann - self.ApplyDirichlet(K[self.mechanical_dofs,:][:,self.mechanical_dofs],
                    # nodal_forces_mech, self.mech_out, self.mech_in, self.applied_dirichlet_mech, LoadFactor=LoadFactor)

            nodal_forces_mech = np.zeros_like(nodal_forces_mech)
            residual_mech = residual_mech_from_elec + residual_mech_neumann - \
                self.ApplyDirichlet(K[self.mechanical_dofs,:][:,self.mechanical_dofs],
                nodal_forces_mech, self.mech_out, self.mech_in, self.applied_dirichlet_mech, LoadFactor=LoadFactor)

            # SOLVE MECHANICS PROBLEM WITH OLD GEOMETRY (K), AND THE FORCE self.force_up AS A RESIDUAL
            dUm = self.SolveMechanics(mesh, formulation, solver, K, residual_mech, LoadFactor, initial_solution=False)
            # dUm = self.SolveMechanics(mesh, formulation, solver, K, traction_forces_mech, LoadFactor, initial_solution=False)
            # UPDATE GEOMETRY INCREMENTALLY
            Eulerx += dUm

            # UPDATE SOLUTION FOR THE CURRENT LOAD INCREMENT
            TotalDisp[:,:formulation.ndim,Increment] = Eulerx - mesh.points
            TotalDisp[:,-1,Increment] = Eulerp

            K = Assemble(self, function_spaces[0], formulation, mesh, material,
                    Eulerx,Eulerp)[0]

            # LOG REULTS
            self.LogSave(formulation, TotalDisp, Increment)

            print('\nFinished Load increment', Increment, 'in', time()-t_increment, 'seconds')

            # BREAK AT A SPECIFICED LOAD INCREMENT IF ASKED FOR
            if self.break_at_increment != -1 and self.break_at_increment is not None:
                if self.break_at_increment == Increment:
                    if self.break_at_increment < LoadIncrement - 1:
                        print("\nStopping at increment {} as specified\n\n".format(Increment))
                        TotalDisp = TotalDisp[:,:,:Increment]
                        self.number_of_load_increments = Increment
                    break


        return TotalDisp


    def ApplyDirichlet(self, stiffness, F, columns_out, columns_in, AppliedDirichlet, LoadFactor=1., mass=None):
        """AppliedDirichlet is a non-member because it can be external incremental Dirichlet,
            which is currently not implemented as member of BoundaryCondition. F also does not
            correspond to Dirichlet forces, as it can be residual in incrementally linearised
            framework.
        """

        # APPLY DIRICHLET BOUNDARY CONDITIONS
        for i in range(columns_out.shape[0]):
            F = F - LoadFactor*AppliedDirichlet[i]*stiffness.getcol(columns_out[i])

        return F


    def SolveMechanics(self, mesh, formulation, solver, K, residual_mech, LoadFactor, initial_solution=True):
        """ Solves for mechanical variables. This solves the upper row
            of the following system

                [K_uu K_up][U_u] = [F_u]
                [K_pu K_pp][U_p] = [F_p]

            i.e.

                K_uu*U_u = F_u - K_up*U_p


            input:

                K:              [scipy.sparse.csc/csr_matrix] Total electromechanical stiffness
                                Matrix, no boundary conditions applied

        """

        K_uu_b = K[self.columns_in_mech,:][:,self.columns_in_mech]

        if initial_solution:
            F_b = residual_mech[self.mech_in,0]
        else:
            rhs_mech = residual_mech + self.force_up[:,None]
            F_b = rhs_mech[self.mech_in]

        sol = solver.Solve(K_uu_b,-F_b)

        # REARRANGE
        dUm = np.zeros(self.all_mech_dofs.shape[0],dtype=np.float64)
        dUm[self.mech_in] = sol.ravel()
        dUm[self.mech_out] = LoadFactor*self.applied_dirichlet_mech
        dUm = dUm.reshape(int(dUm.shape[0]/formulation.ndim),formulation.ndim)

        return dUm



    def SolveElectrostatics(self, K, residual_electric,formulation,solver,iteration):
        """ Solves for mechanical variables. This solves the lower row
            of the following system

                [K_uu K_up][U_u] = [F_u]
                [K_pu K_pp][U_p] = [F_p]

            i.e.

                K_uu*U_u = F_u - K_up*U_p


            input:

                K:              [scipy.sparse.csc/csr_matrix] Total electromechanical stiffness
                                Matrix, no boundary conditions applied

        """
        K_pp_b = K[self.columns_in_electric,:][:,self.columns_in_electric]
        if iteration == 0:
            rhs_electric = residual_electric + self.force_pu[:,None]
        else:
            rhs_electric = residual_electric

        F_b = rhs_electric[self.electric_in]
        sol = solver.Solve(K_pp_b,-F_b)
        # REARRANGE
        dUe = np.zeros(self.all_electric_dofs.shape[0])
        dUe[self.electric_in] = sol.ravel()

        return dUe


from __future__ import print_function
import os, platform
from time import time
from warnings import warn
import numpy as np
import scipy as sp
from scipy.sparse import issparse, isspmatrix_coo, isspmatrix_csr, isspmatrix_csc
from scipy.sparse.linalg import spsolve, cg, cgs, bicgstab, gmres, lgmres, splu, spilu, LinearOperator, onenormest
from subprocess import call


__all__ = ["LinearSolver"]


class LinearSolver(object):
    """Base class for all linear sparse direct and iterative solvers"""

    def __init__(self,linear_solver="direct", linear_solver_type="umfpack",
        apply_preconditioner=False, preconditioner="smoothed_aggregation",
        iterative_solver_tolerance=1.0e-12, reduce_matrix_bandwidth=False,
        out_of_core=False, geometric_discretisation=None, dont_switch_solver=False):
        """

            input:
                linear_solver:          [str] type of solver either "direct",
                                        "iterative" or "multigrid"

                linear_solver_type      [str] type of direct or linear solver to
                                        use, for instance "umfpack", "superlu" or
                                        "mumps" for direct solvers, or "cg", "gmres"
                                        etc for iterative solvers or "amg" for algebraic
                                        multigrid solver. See WhichSolvers method for
                                        the complete set of available linear solvers

                preconditioner:         [str] either "amg_smoothed_aggregation",
                                        or "ruge_stuben" or "rootnode" for
                                        a preconditioner based on algebraic multigrid
                                        or "incomplete_lu" for scipy's spilu linear
                                        operator

                geometric_discretisation:
                                        [str] type of geometric discretisation used, for
                                        instance for FEM discretisations this would correspond
                                        to "tri", "quad", "tet", "hex" etc

                dont_switch_solver:     Do not switch between solvers automatically


        """

        self.is_sparse = True
        self.solver_type = linear_solver
        self.solver_subtype = linear_solver_type
        self.requires_cuthill_mckee = reduce_matrix_bandwidth
        self.iterative_solver_tolerance = iterative_solver_tolerance
        self.apply_preconditioner = apply_preconditioner
        self.preconditioner_type = preconditioner
        self.out_of_core = False
        self.geometric_discretisation = geometric_discretisation
        self.dont_switch_solver = dont_switch_solver
        self.reuse_factorisation = False
        self.solver_context_manager = None

        self.has_amg_solver = True
        if platform.python_implementation() == "PyPy":
            self.has_amg_solver = False
        else:
            try:
                import pyamg
            except ImportError:
                self.has_amg_solver = False

        self.has_umfpack = True
        try:
            from scikits.umfpack import spsolve
        except ImportError:
            self.has_umfpack = False

        self.has_mumps = False
        try:
            from mumps.mumps_context import MUMPSContext
            self.has_mumps = True
        except ImportError:
            self.has_mumps = False

        self.has_pardiso = False
        try:
            import pypardiso
            self.has_pardiso = True
        except ImportError:
            self.has_pardiso = False

        self.switcher_message = False

        # self.analysis_type = "static"
        # self.analysis_nature = "linear"

    def SetSolver(self,linear_solver="direct", linear_solver_type="umfpack",
        apply_preconditioner=False, preconditioner="amg_smoothed_aggregation",
        iterative_solver_tolerance=1.0e-12, reduce_matrix_bandwidth=False,
        geometric_discretisation=None):
        """

            input:
                linear_solver:          [str] type of solver either "direct",
                                        "iterative" or "multigrid"

                linear_solver_type      [str] type of direct or linear solver to
                                        use, for instance "umfpack", "superlu" or
                                        "mumps" for direct solvers, or "cg", "gmres"
                                        etc for iterative solvers or "amg" for algebraic
                                        multigrid solver. See WhichSolvers method for
                                        the complete set of available linear solvers

                preconditioner:         [str] either "amg_smoothed_aggregation" for
                                        a preconditioner based on algebraic multigrid
                                        or "incomplete_lu" for scipy's spilu linear
                                        operator

                geometric_discretisation:
                                        [str] type of geometric discretisation used, for
                                        instance for FEM discretisations this would correspond
                                        to "tri", "quad", "tet", "hex" etc

        """

        self.solver_type = linear_solver
        self.solver_subtype = "umfpack"
        self.iterative_solver_tolerance = iterative_solver_tolerance
        self.apply_preconditioner = apply_preconditioner
        self.requires_cuthill_mckee = reduce_matrix_bandwidth
        self.geometric_discretisation = geometric_discretisation


    @property
    def WhichLinearSolver(self):
        return self.solver_type, solver_subtype

    @property
    def WhichLinearSolvers(self):
        return {"direct":["superlu","umfpack","mumps"],
            "iterative":["cg","bicg","cgstab","bicgstab"
            "gmres","lgmres"],"multigrid":["amg"]}


    def GetPreconditioner(self,A, type="amg_smoothed_aggregation"):
        """Applies a suitable preconditioner to sparse matrix A
            based on algebraic multigrid of incomplete LU/Cholesky factorisation

            input:
                A:                      [csc_matrix or csc_matrix]
                type:                   [str] either "amg_smoothed_aggregation" for
                                        a preconditioner based on algebraic multigrid
                                        or "incomplete_lu" for scipy's spilu linear
                                        operator

            returns:                    A preconditioner that can be used in conjunction
                                        with scipy's sparse linear iterative solvers
                                        (the M keyword in scipy's iterative solver)
        """

        if not (isspmatrix_csc(A) or isspmatrix_csr(A)):
            raise TypeError("Matrix must be in CSC or CSR sparse format for preconditioning")

        ml = smoothed_aggregation_solver(A)
        return ml.aspreconditioner()


    def GetCuthillMcKeePermutation(self,A):
        """Applies Cuthill-Mckee permutation to reduce the sparse matrix bandwidth

            input:
                A:                    [csc_matrix or csr_matrix]

            returns:
                perm:                 [1D array] of permutation such that A[perm,:][:,perm]
                                      has its non-zero elements closer to the diagonal
        """

        if not (isspmatrix_csc(A) or isspmatrix_csr(A)):
            raise TypeError("Matrix must be in CSC or CSR sparse format "
                "for Cuthill-McKee permutation")

        if int(sp.__version__.split('.')[1]) >= 15:
            from scipy.sparse.csgraph import reverse_cuthill_mckee
            perm = reverse_cuthill_mckee(A)
        else:
            from Florence.Tensor import symrcm
            perm = symrcm(A)

        return perm


    def SparsityPattern(self,A):
        import matplotlib.pyplot as plt
        plt.spy(A)
        plt.grid('on')
        plt.show()


    def Solve(self, A, b, reuse_factorisation=False):
        """Solves the linear system of equations"""

        self.reuse_factorisation = reuse_factorisation
        if self.solver_type != "direct" and self.reuse_factorisation is True:
            warn("Re-using factorisation for non-direct solvers is not possible. The pre-conditioner is going to be reused instead")

        # DECIDE IF THE SOLVER TYPE IS APPROPRIATE FOR THE PROBLEM
        if self.switcher_message is False and self.dont_switch_solver is False:
            # PREFER PARDISO OR MUMPS IF AVAILABLE
            if self.has_pardiso:
                self.solver_type = "direct"
                self.solver_subtype = "pardiso"
            elif self.has_mumps:
                self.solver_type = "direct"
                self.solver_subtype = "mumps"
            elif b.shape[0] > 100000 and self.has_amg_solver:
                self.solver_type = "multigrid"
                self.solver_subtype = "amg"
                print('Large system of equations. Switching to algebraic multigrid solver')
                self.switcher_message = True
            # elif mesh.points.shape[0]*MainData.nvar > 50000 and MainData.C < 4:
                # self.solver_type = "direct"
                # self.solver_subtype = "MUMPS"
                # print 'Large system of equations. Switching to MUMPS solver'
            elif b.shape[0] > 70000 and self.geometric_discretisation=="hex" and self.has_amg_solver:
                self.solver_type = "multigrid"
                self.solver_subtype = "amg"
                print('Large system of equations. Switching to algebraic multigrid solver')
                self.switcher_message = True
            else:
                self.solver_type = "direct"
                self.solver_subtype = "umfpack"


        if self.solver_type == 'direct':
            # CALL DIRECT SOLVER
            if self.solver_subtype=='umfpack' and self.has_umfpack:
                if A.dtype != np.float64:
                    A = A.astype(np.float64)

                if self.solver_context_manager is None:
                    if self.reuse_factorisation is False:
                        sol = spsolve(A,b,permc_spec='MMD_AT_PLUS_A',use_umfpack=True)
                        # from scikits import umfpack
                        # sol = umfpack.spsolve(A, b)
                    else:
                        from scikits import umfpack
                        lu = umfpack.splu(A)
                        sol = lu.solve(b)
                        self.solver_context_manager = lu
                else:
                    sol = self.solver_context_manager.solve(b)


            elif self.solver_subtype=='mumps' and self.has_mumps:

                from mumps.mumps_context import MUMPSContext
                t_solve = time()
                A = A.tocoo()
                # False means non-symmetric - Do not change it to True. True means symmetric pos def
                # which is not the case for electromechanics
                if self.solver_context_manager is None:
                    context = MUMPSContext((A.shape[0], A.row, A.col, A.data, False), verbose=False)
                    context.analyze()
                    context.factorize()
                    sol = context.solve(rhs=b)

                    if self.reuse_factorisation:
                        self.solver_context_manager = context
                else:
                    sol = self.solver_context_manager.solve(rhs=b)

                print("MUMPS solver time is {}".format(time() - t_solve))

                return sol


            elif self.solver_subtype == "pardiso" and self.has_pardiso:
                # NOTE THAT THIS PARDISO SOLVER AUTOMATICALLY SAVES THE RIGHT FACTORISATION
                import pypardiso
                A = A.tocsr()
                t_solve = time()
                sol = pypardiso.spsolve(A,b)
                print("Pardiso solver time is {}".format(time() - t_solve))

            else:
                # FOR 'super_lu'
                if A.dtype != np.float64:
                    A = A.astype(np.float64)
                A = A.tocsc()

                if self.solver_context_manager is None:
                    if self.reuse_factorisation is False:
                        sol = spsolve(A,b,permc_spec='MMD_AT_PLUS_A',use_umfpack=True)
                    else:
                        lu = splu(A)
                        sol = lu.solve(b)
                        self.solver_context_manager = lu
                else:
                    sol = self.solver_context_manager.solve(b)



        elif self.solver_type == "iterative":
            # CALL ITERATIVE SOLVER
            # sol = bicgstab(A,b,tol=tol)[0]
            # sol = gmres(A,b,tol=tol)[0]
            sol = cg(A,b,tol=self.iterative_solver_tolerance)[0]

            # PRECONDITIONED ITERATIVE SOLVER - CHECK
            # P = spilu(A, drop_tol=1e-5)
            # M_x = lambda x: P.solve(x)
            # n = A.shape[0]
            # m = A.shape[1]
            # M = LinearOperator((n * m, n * m), M_x)
            # sol = lgmres(A, b, tol=1e-4, M=M)[0]

        elif self.solver_type == "multigrid":
            if self.has_amg_solver is False:
                raise ImportError('A multigrid solver was not found')
            from pyamg import ruge_stuben_solver, rootnode_solver, smoothed_aggregation_solver

            if A.dtype != b.dtype:
                # DOWN-CAST
                b = b.astype(A.dtype)

            if not isspmatrix_csr(A):
                A = A.tocsr()

            t_solve = time()
            # AMG METHOD
            if self.solver_context_manager is None:
                if self.preconditioner_type=="smoothed_aggregation":
                    # EXPLICIT CALL TO KYROLOV SOLVERS WITH AMG PRECONDITIONER
                    # THIS IS TYPICALLY FASTER BUT THE TOLERANCE NEED TO BE SMALLER, TYPICALLY 1e-10
                    # GMRES IS TYPICALLY THE FASTEST
                    ml = smoothed_aggregation_solver(A)
                    M = ml.aspreconditioner()
                    if self.iterative_solver_tolerance > 1e-9:
                        self.iterative_solver_tolerance = 1e-10
                    # sol, info = bicgstab(A, b, M=M, tol=self.iterative_solver_tolerance)
                    # sol, info = cgs(A, b, M=M, tol=self.iterative_solver_tolerance)
                    sol, info = gmres(A, b, M=M, tol=self.iterative_solver_tolerance)
                elif self.preconditioner_type == "ruge_stuben":
                    M = ruge_stuben_solver(A)
                    sol = M.solve(b,tol=self.iterative_solver_tolerance)
                elif self.preconditioner_type == "rootnode":
                    # EXPLICIT CALL TO KYROLOV SOLVERS WITH AMG PRECONDITIONER
                    # ml = rootnode_solver(A, smooth=('energy', {'degree':2}), strength='evolution' )
                    # M = ml.aspreconditioner(cycle='V')
                    ml = rootnode_solver(A)
                    M = ml.aspreconditioner()
                    if self.iterative_solver_tolerance > 1e-9:
                        self.iterative_solver_tolerance = 1e-10
                    sol, info = gmres(A, b, M=M, tol=self.iterative_solver_tolerance)


                if self.reuse_factorisation:
                    self.solver_context_manager = M

            else:
                M = self.solver_context_manager
                if self.preconditioner_type=="smoothed_aggregation":
                    sol, info = gmres(A, b, M=M, tol=self.iterative_solver_tolerance)
                elif self.preconditioner_type == "ruge_stuben":
                    sol = M.solve(b,tol=self.iterative_solver_tolerance)
                elif self.preconditioner_type == "rootnode":
                    sol, info = gmres(A, b, M=M, tol=self.iterative_solver_tolerance)


            print("AMG solver time is {}".format(time() - t_solve))

        return sol


    def GetConditionNumber(self,A):
        self.matrix_condition_number = onenormest(K_b)
        return self.matrix_condition_number


    def CleanUp(self):
        import gc
        del self.solver_context_manager
        gc.collect()
        self.solver_context_manager = None



import numpy as np

def GetBases(C,z):
    from Florence.FunctionSpace.OneDimensional.Line import LagrangeGaussLobatto, Lagrange
    # Get basis at all integration points - every column corresponds to a Gauss point
    Basis = np.zeros((C+2,z.shape[0])); dBasis = np.copy(Basis)
    for i in range(0,z.shape[0]):
        # Basis[0:,i], dBasis[0:,i], _ = Lagrange(C,z[i])
        Basis[0:,i], dBasis[0:,i], _ = LagrangeGaussLobatto(C,z[i])

    return Basis, dBasis


import numpy as np


class FunctionSpace(object):
    """Base class for all interpolation functions for finite element
        and boundary element analyses
    """

    def __init__(self, mesh, quadrature=None, p=1, bases_type="nodal", bases_kind="CG",
        evaluate_at_nodes=False, equally_spaced=False, use_optimal_quadrature=False):
        """

            input:
                mesh:                       [Mesh] an instance of class Mesh
                p:                          [int] degree of interpolation bases
                bases_type:                 [str] type of interpolation bases,
                                            either "nodal" for higher order of
                                            "modal" for hierarchical bases
                bases_kind:                 [str] kind of interpolation bases,
                                            either "CG" for continuous Galerkin
                                            or "DG" for discontinuous Galerkin
                equally_spaced:             Only applicable to nodal bases functions.
                                            Wether the position of nodes in the
                                            isoparametric domain should be equally spaced
                                            or not
        """

        # from Florence import QuadratureRule
        from Florence.FunctionSpace.GetBases import GetBases1D, GetBases2D, GetBases3D, GetBoundaryBases, GetBasesAtNodes

        ndim = mesh.InferSpatialDimension()
        self.element_type = mesh.element_type
        self.ndim = ndim

        QuadratureOpt=3
        is_flattened = False
        if use_optimal_quadrature:
            QuadratureOpt=3 # always this for tris/tets
            is_flattened = True

        # mesh.InferBoundaryElementType()
        C = p - 1
        if mesh.InferPolynomialDegree() - 1 != C:
            raise ValueError("Function space of the polynomial does not match element type")

        if evaluate_at_nodes is False:
            if quadrature is None:
                raise ValueError("Function space requires a quadrature rule")
            # quadrature = QuadratureRule(optimal=QuadratureOpt, norder=norder, mesh_type=mesh.element_type)
            z = quadrature.points
            w = quadrature.weights

        if evaluate_at_nodes is False:
            if mesh.element_type == "tet" or mesh.element_type == "hex":
                # GET BASES AT ALL INTEGRATION POINTS (VOLUME)
                Domain = GetBases3D(C,quadrature,mesh.element_type,equally_spaced=equally_spaced,
                    is_flattened=use_optimal_quadrature)
            elif mesh.element_type == 'tri' or mesh.element_type == 'quad':
                # GET BASES AT ALL INTEGRATION POINTS (AREA)
                Domain = GetBases2D(C,quadrature,mesh.element_type,equally_spaced=equally_spaced,
                    is_flattened=use_optimal_quadrature)
            elif mesh.element_type == 'line':
                # GET BASES AT ALL INTEGRATION POINTS (LINE)
                Domain = GetBases1D(C,quadrature,mesh.element_type,equally_spaced=equally_spaced)
            # Boundary = []
            # # PUT QUADRATURE NONE AS BASES QUADRATURE RULE IS DIFFERENT
            # Boundary = GetBoundaryBases(C,None,mesh.boundary_element_type,equally_spaced=equally_spaced,
                # is_flattened=use_optimal_quadrature)
        else:
            Domain = GetBasesAtNodes(C,quadrature,mesh.element_type,equally_spaced=equally_spaced)
            w = Domain.w
            # Boundary = []


        # COMPUTING GRADIENTS AND JACOBIAN A PRIORI FOR ALL INTEGRATION POINTS
        ############################################################################
        Domain.Jm = []; Domain.AllGauss=[]
        if mesh.element_type == 'hex':
            if is_flattened is False:
                Domain.Jm = np.zeros((ndim,Domain.Bases.shape[0],w.shape[0]**ndim))
                Domain.AllGauss = np.zeros((w.shape[0]**ndim,1))
                counter = 0
                for g1 in range(w.shape[0]):
                    for g2 in range(w.shape[0]):
                        for g3 in range(w.shape[0]):
                            # GRADIENT TENSOR IN PARENT ELEMENT [\nabla_\varepsilon (N)]
                            Domain.Jm[0,:,counter] = Domain.gBasesx[:,counter]
                            Domain.Jm[1,:,counter] = Domain.gBasesy[:,counter]
                            Domain.Jm[2,:,counter] = Domain.gBasesz[:,counter]

                            Domain.AllGauss[counter,0] = w[g1]*w[g2]*w[g3]

                            counter +=1
            else:
                Domain.Jm = np.zeros((ndim,Domain.Bases.shape[0],w.shape[0]))
                Domain.AllGauss = np.zeros((w.shape[0],1))
                for counter in range(0,w.shape[0]):
                    # GRADIENT TENSOR IN PARENT ELEMENT [\nabla_\varepsilon (N)]
                    Domain.Jm[0,:,counter] = Domain.gBasesx[:,counter]
                    Domain.Jm[1,:,counter] = Domain.gBasesy[:,counter]
                    Domain.Jm[2,:,counter] = Domain.gBasesz[:,counter]

                    Domain.AllGauss[counter,0] = w[counter]

        elif mesh.element_type == 'quad':
            if is_flattened is False:
                Domain.Jm = np.zeros((ndim,Domain.Bases.shape[0],w.shape[0]**ndim))
                Domain.AllGauss = np.zeros((w.shape[0]**ndim,1))
                counter = 0
                for g1 in range(w.shape[0]):
                    for g2 in range(w.shape[0]):
                        # GRADIENT TENSOR IN PARENT ELEMENT [\nabla_\varepsilon (N)]
                        Domain.Jm[0,:,counter] = Domain.gBasesx[:,counter]
                        Domain.Jm[1,:,counter] = Domain.gBasesy[:,counter]

                        Domain.AllGauss[counter,0] = w[g1]*w[g2]
                        counter +=1
            else:
                Domain.Jm = np.zeros((ndim,Domain.Bases.shape[0],w.shape[0]))
                Domain.AllGauss = np.zeros((w.shape[0],1))
                for counter in range(0,w.shape[0]):
                    # GRADIENT TENSOR IN PARENT ELEMENT [\nabla_\varepsilon (N)]
                    Domain.Jm[0,:,counter] = Domain.gBasesx[:,counter]
                    Domain.Jm[1,:,counter] = Domain.gBasesy[:,counter]

                    Domain.AllGauss[counter,0] = w[counter]

        elif mesh.element_type == 'tet':
            Domain.Jm = np.zeros((ndim,Domain.Bases.shape[0],w.shape[0]))
            Domain.AllGauss = np.zeros((w.shape[0],1))
            for counter in range(0,w.shape[0]):
                # GRADIENT TENSOR IN PARENT ELEMENT [\nabla_\varepsilon (N)]
                Domain.Jm[0,:,counter] = Domain.gBasesx[:,counter]
                Domain.Jm[1,:,counter] = Domain.gBasesy[:,counter]
                Domain.Jm[2,:,counter] = Domain.gBasesz[:,counter]

                Domain.AllGauss[counter,0] = w[counter]

        elif mesh.element_type == 'tri':
            Domain.Jm = np.zeros((ndim,Domain.Bases.shape[0],w.shape[0]))
            Domain.AllGauss = np.zeros((w.shape[0],1))
            for counter in range(0,w.shape[0]):
                # GRADIENT TENSOR IN PARENT ELEMENT [\nabla_\varepsilon (N)]
                Domain.Jm[0,:,counter] = Domain.gBasesx[:,counter]
                Domain.Jm[1,:,counter] = Domain.gBasesy[:,counter]

                Domain.AllGauss[counter,0] = w[counter]

        elif mesh.element_type == "line":
            Domain.Jm = np.zeros((ndim,Domain.Bases.shape[0],w.shape[0]))
            Domain.AllGauss = np.zeros((w.shape[0],1))
            for counter in range(0,w.shape[0]):
                # GRADIENT TENSOR IN PARENT ELEMENT [\nabla_\varepsilon (N)]
                Domain.Jm[0,:,counter] = Domain.gBasesx[:,counter]

                Domain.AllGauss[counter,0] = w[counter]


        self.Jm = Domain.Jm
        self.AllGauss = Domain.AllGauss
        self.Bases = Domain.Bases
        self.gBasesx = Domain.gBasesx
        self.gBasesy = Domain.gBasesy
        self.gBasesz = Domain.gBasesz



from __future__ import division, print_function
from warnings import warn
import numpy as np
from Florence.Tensor import makezero, makezero3d, unique2d


class GeometricPath(object):
    """Construct a geometric path for mesh extrusions
    """

    def __init__(self):
        pass

    def ComputeExtrusion(self,base_mesh):
        pass





class GeometricLine(GeometricPath):

    def __init__(self, start=(0.,0.,0.), end=(1.,2.,3)):
        """Constructs a line given the start and ending points 
        """

        self.start = np.array(start)
        self.end = np.array(end)
        self.length = np.linalg.norm(self.start - self.end)

    
    def ComputeExtrusion(self, nlong=10):
        """Computes extrusion of base_mesh along the self (line)
            using equal spacing 

            input:
                nlong:                      [int] number of discretisation along
                                            the arc
            returns:
                points:                     [1D array] of discretisation along arc
        """


        from numpy.linalg import norm

        npoints = nlong + 1
        x = np.linspace(self.start[0],self.end[0],npoints)
        y = np.linspace(self.start[1],self.end[1],npoints)
        z = np.linspace(self.start[2],self.end[2],npoints)

        points = np.array([x,y,z]).T.copy()
        # IF ALL THE Z-COORDIDATES ARE ZERO, THEN EXTRUSION NOT POSSIBLE
        if np.allclose(points[:,2],0.0):
            warn("Third dimension is zero. Extrusion will fail, unless axes are swapped")

        return points






class GeometricArc(GeometricPath):

    def __init__(self, center=(0.,0.,0.), start=(-1.,-2.,-3.), end=(3.,2.,1.)):
        """Constructs an arc given the center of the arc and start
            and ending points of the arc
        """
        
        from numpy.linalg import norm

        self.center = np.array(center)
        self.start = np.array(start)
        self.end = np.array(end)

        # COMPUTE ARC'S RADIUS
        self.radius = norm(self.start - self.center)
        if not np.isclose(self.radius, norm(self.end - self.center)):
            raise ValueError("An arc could not be constructed with the given points")

        # COMPUTE ANGLE OF ARC
        line0 = self.start - self.center
        line1 = self.end   - self.center
        self.angle = np.arccos(norm(line0*line1)/norm(line0)/norm(line1))
        # self.angle = np.arccos(np.abs(np.dot(line0,line1))/norm(line0)/norm(line1))


    def Construct(self, center=(0.,0.), start=(-2.,2.), end=(2.,2.)):
        """Constructs an arc given the center of the arc and start
            and ending points of the arc
        """
        self.init(center=center, start=start, end=end)


    def SetAngles(self, start_angle_phi=None, end_angle_phi=None,
        start_angle_theta=None, end_angle_theta=None):
        """Set the arc angle in radians
        """

        if start_angle_phi is not None:
            self.start_angle_phi = start_angle_phi
        if end_angle_phi is not None:
            self.end_angle_phi = end_angle_phi
        if start_angle_theta is not None:
            self.start_angle_theta = start_angle_theta
        if end_angle_theta is not None:
            self.end_angle_theta = end_angle_theta


    def ComputeExtrusion(self, nlong=10):
        """Computes extrusion of a base mesh along an arc

            input:
                nlong:                      [int] number of discretisation along
                                            the arc
            returns:
                points:                     [1D array] of discretisation along arc
        """

        from numpy.linalg import norm

        npoints = nlong + 1

        line0 = self.start - self.center
        line1 = self.end   - self.center

        # CHECK IF ALL THE POINTS LIE ON A LINE
        if np.isclose(np.abs(self.angle),0.0) or \
            np.isclose(np.abs(self.angle),np.pi) or \
            np.isclose(np.abs(self.angle),2.*np.pi):
            self.angle += 1e-08
            warn("The arc points are nearly colinear")

        t = np.linspace(0,self.angle,npoints+1)

        points = (np.einsum("i,j",np.sin(self.angle - t),line0) + \
            np.einsum("i,j",np.sin(t),line1))/np.sin(self.angle) + self.center
        makezero(points)

        # IF ALL THE Z-COORDIDATES ARE ZERO, THEN EXTRUSION NOT POSSIBLE
        if np.allclose(points[:,2],0.0):
            warn("Third dimension is zero. Extrusion will fail, unless axes are swapped")

        return points




        # # print(tx*self.angle)
        # # FIND NORMAL TO THE PLANE OF ARC
        # normal_to_arc_plane = np.cross(self.start - self.center, self.end - self.center)
        # # FIND TWO ORTHOGONAL VECTORS IN THE PLANE OF ARC
        # X = self.start/norm(self.start)
        # Y = np.cross(normal_to_arc_plane, self.start) / norm(np.cross(normal_to_arc_plane, self.start))
        # # print(normal_to_arc_plane, self.start)
        # # t = np.linspace(0, self.angle, npoints+1)
        # t = np.linspace(0., self.angle, npoints+1)
        # # t = np.linspace(0, self.angle+0.2284, npoints+1)
        # # x = norm(np.dot(self.start - self.center, self.end - self.center))
        # # print(np.arccos(x))
        # # print(self.angle)

        # points = np.einsum("i,j",self.radius*np.cos(t),X) + np.einsum("i,j",self.radius*np.sin(t),Y)
        # # points = np.einsum("i,j",self.radius*np.sin(t),X) + np.einsum("i,j",self.radius*np.cos(t),Y)
        # # y = np.einsum("i,j",self.radius*np.sin(t),Y)
        # # print(X,Y,t)
        # print(points)
        # # print(np.dot(X,Y))

        # # t= np.linspace(-np.pi/2,np.pi,100)
        # # X = np.array([0,1])
        # # Y = np.array([1,0])
        # # # points = np.einsum("i,j",self.radius*np.cos(t),X) + np.einsum("i,j",self.radius*np.sin(t),Y)
        # # points = np.einsum("i,j",self.radius*np.sin(t),X) + np.einsum("i,j",self.radius*np.cos(t),Y)
        # # makezero(points)
        # # print(points)
        # exit()

        # import matplotlib.pyplot as plt
        # plt.plot(points[:,0],points[:,1],'ro')
        # plt.axis('equal')
        # plt.show()

        # import os
        # os.environ['ETS_TOOLKIT'] = 'qt4'
        # from mayavi import mlab

        # figure = mlab.figure(bgcolor=(1,1,1),fgcolor=(1,1,1),size=(800,600))
        # mlab.plot3d(points[:,0],points[:,1],points[:,2])
        # mlab.show()
        # exit()

        # print(self.start_angle, self.end_angle)
        # x = self.radius*np.cos(t)
        # y = self.radius*np.sin(t)


        # points = np.array([x,y]).T.copy()
        # makezero(points)

        # t = np.linspace(0,40.,npoints+1)


import numpy as np
from numpy import einsum
from Florence.Tensor import trace, Voigt
from .MaterialBase import Material
from Florence.LegendreTransform import LegendreTransform


class MooneyRivlin(Material):
    """The fundamental MooneyRivlin model from Gil and Ortigosa et. al.

        W_mn(C) = u1*C:I+u2*G:I - 2*(u1+2*u2)*lnJ + lamb/2*(J-1)**2

    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(MooneyRivlin, self).__init__(mtype, ndim, **kwargs)
        self.is_transversely_isotropic = False
        self.energy_type = "internal_energy"
        self.nature = "nonlinear"
        self.fields = "mechanics"

        if self.ndim==3:
            self.H_VoigtSize = 6
        elif self.ndim==2:
            self.H_VoigtSize = 3

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = True
        # self.has_low_level_dispatcher = False

    def KineticMeasures(self,F,ElectricFieldx=0, elem=0):
        from Florence.MaterialLibrary.LLDispatch._MooneyRivlin_ import KineticMeasures
        return KineticMeasures(self,F)


    def Hessian(self,StrainTensors,ElectricDisplacementx,elem=0,gcounter=0):

        mu1 = self.mu1
        mu2 = self.mu2
        lamb = self.lamb

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]

        C_Voigt = 2.*mu2/J*(2*einsum('ij,kl',b,b) - einsum('ik,jl',b,b) - einsum('il,jk',b,b)) +\
            2.*(mu1+2.*mu2)/J*( einsum("ik,jl",I,I)+einsum("il,jk",I,I) ) + \
            lamb*(2.*J-1.)*einsum("ij,kl",I,I) - lamb*(J-1)*( einsum("ik,jl",I,I)+einsum("il,jk",I,I) )

        C_Voigt = Voigt(C_Voigt,1)


        return C_Voigt



    def CauchyStress(self,StrainTensors,ElectricDisplacementx,elem=0,gcounter=0):

        mu1 = self.mu1
        mu2 = self.mu2
        lamb = self.lamb

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]

        trb = trace(b)
        if self.ndim==2:
            trb +=1.

        sigma = 2.0*mu1/J*b + \
            2.0*mu2/J*(trb*b - np.dot(b,b)) -\
            2.0*(mu1+2*mu2)/J*I +\
            lamb*(J-1)*I

        return sigma


    def InternalEnergy(self,StrainTensors,elem=0,gcounter=0):

        mu1 = self.mu1
        mu2 = self.mu2
        lamb = self.lamb

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        F = StrainTensors['F'][gcounter]
        H = J*np.linalg.inv(F).T
        C = np.dot(F.T,F)
        G = np.dot(H.T,H)

        energy  = mu1*(einsum('ij,ij',C,I) - 3.) + mu2*(einsum('ij,ij',G,I) - 3.) - 2.*(mu1+2.*mu2)*np.log(J) + lamb/2.*(J-1)**2

        return energy



from __future__ import print_function
import numpy as np

def FEMDebugger(function_space, quadrature_rule, mesh):
    """Generic florence debugger for finite elements"""


    Domain = function_space
    Quadrature = quadrature_rule

    ndim = mesh.points.shape[1]

    tick = u'\u2713'.encode('utf8')+' : ' 
    cross = u'\u2717'.encode('utf8')+' : '

    # CHECK GAUSS POINTS
    sum_bases = np.sum(Domain.Bases,axis=0)
    sum_gbasesx = np.sum(Domain.gBasesx,axis=0)
    sum_gbasesy = np.sum(Domain.gBasesy,axis=0)
    sum_gbasesz = np.array([])
    if ndim==3:
        sum_gbasesz = np.sum(Domain.gBasesz,axis=0)
    sum_gauss_weights = np.sum(Quadrature.weights)
    
    isones = np.ones_like(sum_bases)
    iszeros = np.zeros_like(sum_gbasesx)
    
    # CHECK QUADRATURE RULE
    element_area = {'tri':2.0,'tet':4.0/3.0,'quad':4,'hex':8}
    fem_element_area = {'tri':sum_gauss_weights,'tet':sum_gauss_weights,
        'quad':sum_gauss_weights**2,'hex':sum_gauss_weights**3}
    if np.isclose(element_area[mesh.element_type],fem_element_area[mesh.element_type]):
        print(tick, 'Summation of all quadrature weights equals area of the parent element')
    else:
        print(cross, 'Summation of all quadrature weights does NOT equal area of the parent element')

    # CHECK BASIS FUNCTIONS
    if np.allclose(sum_bases,isones):
        print(tick, 'Summation of all interpolation functions at every Gauss point is one')
    else:
        print(cross, 'Summation of all interpolation functions at every Gauss point is NOT one')
    if np.allclose(sum_gbasesx,iszeros):
        print(tick, 'Summation of X-gradient of interpolation functions at every Gauss point is zero')
    else:
        print(cross, 'Summation of X-gradient of interpolation functions at every Gauss point is NOT zero')
    if np.allclose(sum_gbasesy,iszeros):
        print(tick, 'Summation of Y-gradient of interpolation functions at every Gauss point is zero')
    else:
        print(cross, 'Summation of Y-gradient of interpolation functions at every Gauss point is NOT zero')
    if ndim == 3:
        if np.allclose(sum_gbasesx,iszeros):
            print(tick, 'Summation of Z-gradient of interpolation functions at every Gauss point is zero')
        else:
            print(cross, 'Summation of Z-gradient of interpolation functions at every Gauss point is NOT zero')

    # CHECK MESH
    mesh.CheckNodeNumbering()


def NonlinearMechanicsDebugger(mesh,formulation,material):

    tick = u'\u2713'.encode('utf8')+' : ' 
    cross = u'\u2717'.encode('utf8')+' : '
    
    from Florence.FiniteElements.ElementalMatrices.KinematicMeasures import KinematicMeasures
    StrainTensors = KinematicMeasures(np.asarray([np.eye(self.ndim,self.ndim)]*2),"nonlinear")
    stress = material.CauchyStress(StrainTensors,np.zeros((self.ndim,1)))

    if formulation.fields == "electro-mechanics":
        D = material.ElectricDisplacementx(StrainTensors,np.zeros((self.ndim,1)))

    if not np.allclose(stress,0.0): # and not formulation.has_prestress
        print(tick, 'Cauchy stress at the origin is zero')
    else:
        print(cross, 'Cauchy stress at the origin is NOT zero')

    if formulation.fields == "electro-mechanics":
        if not np.allclose(D,0.0):
            print(tick, 'Electric displacement at the origin is zero')
        else:
            print(cross, 'Electric displacement at the origin is NOT zero')


#cython: profile=False
#cython: infer_types=True
#cython: boundscheck=False
#cython: wraparound=False
#cython: nonecheck=False

import numpy as np
cimport numpy as np


ctypedef double Real


cdef extern from "_KinematicMeasures_.h":
    void KinematicMeasures(Real *SpatialGradient_, Real *F_, Real *detJ, const Real *Jm_,
        const Real *AllGauss_, const Real *LagrangeElemCoords_, const Real *EulerElemCoords_,
        int ngauss, int ndim, int nodeperelem, int update) 


def _KinematicMeasures_(np.ndarray[Real, ndim=3, mode='c'] Jm,
    np.ndarray[Real, ndim=1] AllGauss,
    np.ndarray[Real, ndim=2, mode='c'] LagrangeElemCoords, 
    np.ndarray[Real, ndim=2, mode='c'] EulerElemCoords, requires_geometry_update):
    
    cdef int ndim = Jm.shape[0]
    cdef int nodeperelem = Jm.shape[1]
    cdef int ngauss = Jm.shape[2]
    cdef int update = int(requires_geometry_update)

    cdef np.ndarray[Real, ndim=3, mode='c'] F = np.zeros((ngauss,ndim,ndim),dtype=np.float64)
    cdef np.ndarray[Real, ndim=3, mode='c'] SpatialGradient = np.zeros((ngauss,nodeperelem,ndim),dtype=np.float64)
    cdef np.ndarray[Real, ndim=1] detJ = np.zeros(ngauss,dtype=np.float64)

    KinematicMeasures(&SpatialGradient[0,0,0], &F[0,0,0], &detJ[0], &Jm[0,0,0], &AllGauss[0],
        &LagrangeElemCoords[0,0], &EulerElemCoords[0,0], ngauss, ndim, nodeperelem, update)

    return SpatialGradient, F, detJ


from .ImplicitStructuralDynamicIntegrator import LinearImplicitStructuralDynamicIntegrator
from .ImplicitStructuralDynamicIntegrator import NonlinearImplicitStructuralDynamicIntegrator
from .ExplicitStructuralDynamicIntegrator import ExplicitStructuralDynamicIntegrator